#include <stdbool.h>
#include <util/delay.h>     /* for _delay_ms() */
#include "usbdrv.h"

#include <motelib/system.h>
#include <motelib/radio.h>
#include <motelib/uart.h>
#include <motelib/led.h>

enum
{
    MSG_MOVE_X,
    MSG_MOVE_Y,
    MSG_TOGGLE_BTN,
};

#define RECEIVER_NODE_ADDR   0

// Mouse with 3 buttons, middle button is wheel
PROGMEM const char usbHidReportDescriptor[52] = 
{
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x02,                    // USAGE (Mouse)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x09, 0x01,                    //   USAGE (Pointer)
    0xA1, 0x00,                    //   COLLECTION (Physical)
    0x05, 0x09,                    //     USAGE_PAGE (Button)
    0x19, 0x01,                    //     USAGE_MINIMUM
    0x29, 0x03,                    //     USAGE_MAXIMUM
    0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
    0x95, 0x03,                    //     REPORT_COUNT (3)
    0x75, 0x01,                    //     REPORT_SIZE (1)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0x95, 0x01,                    //     REPORT_COUNT (1)
    0x75, 0x05,                    //     REPORT_SIZE (5) -- for padding
    0x81, 0x03,                    //     INPUT (Const,Var,Abs)
    0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x09, 0x38,                    //     USAGE (Wheel)
    0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
    0x25, 0x7F,                    //     LOGICAL_MAXIMUM (127)
    0x75, 0x08,                    //     REPORT_SIZE (8)
    0x95, 0x03,                    //     REPORT_COUNT (3)
    0x81, 0x06,                    //     INPUT (Data,Var,Rel)
    0xC0,                          //   END_COLLECTION
    0xC0,                          // END COLLECTION
};

typedef struct
{
    uint8_t  buttons;
    int8_t   dx,dy;
    int8_t   dwheel;
} report_t;

static report_t reportBuffer;
static uint8_t  idleRate;   /* repeat rate for keyboards, never used for mice */
static bool newRequest;

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
    usbRequest_t *rq = (void *)data;

    /* class request type */
    if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
    {
        if(rq->bRequest == USBRQ_HID_GET_REPORT)
        {
            usbMsgPtr = (void *)&reportBuffer;
            return sizeof(reportBuffer);
        }
        else if (rq->bRequest == USBRQ_HID_GET_IDLE)
        {
            usbMsgPtr = &idleRate;
            return 1;
        }
        else if (rq->bRequest == USBRQ_HID_SET_IDLE)
        {
            idleRate = rq->wValue.bytes[1];
        }
    }
    else
    {
        /* no vendor specific requests implemented */
    }
    return 0;   /* default for not implemented requests: return no data back to host */
}

void loop()
{
    usbPoll();

    if (usbInterruptIsReady() && newRequest)
    {
        ledToggle(0);
        usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
        newRequest = false;
    }
}

void rxHandler(Address source, MessageType type, void *msg, uint8_t len)
{
    int8_t *data = (int8_t*)msg;

    reportBuffer.dx = 0;
    reportBuffer.dy = 0;
    reportBuffer.dwheel = 0;

    newRequest = true;

    if (type == MSG_MOVE_X)
    {
        reportBuffer.dx = data[0];
    }
    else if (type == MSG_MOVE_Y)
    {
        reportBuffer.dy = data[0];
    }
    else if (type == MSG_TOGGLE_BTN)
    {
        uint8_t btn = data[0];
        reportBuffer.buttons ^= (1<<btn);
    }
    else
        newRequest = false;
}

void uartRx(uint8_t markerFound)
{
    uint8_t c;
    uint8_t type;
    int8_t data;

    type = 0xff;
    c = uartReadByte();

    if (c == 'h')
    {
        data = -5;
        type = MSG_MOVE_X;
    }
    else if (c == 'j')
    {
        data = 5;
        type = MSG_MOVE_Y;
    }
    else if (c == 'k')
    {
        data = -5;
        type = MSG_MOVE_Y;
    }
    else if (c == 'l')
    {
        data = 5;
        type = MSG_MOVE_X;
    }
    else if ( (c >= '0') && (c <= '2') )
    {
        data = c - '0';
        type = MSG_TOGGLE_BTN;
    }

    if (type != 0xff)
    {
        radioRequestTx(RECEIVER_NODE_ADDR, type, &data, sizeof(data), NULL);
        ledToggle(2);
    }

    uartWatchInputBytes(1, uartRx);
}

void boot()
{
    if (getAddress() == RECEIVER_NODE_ADDR)
    {
        // This is the receiver node; prepare USB and radio handler
        newRequest = false;
        usbInit();
        usbDeviceDisconnect();
        _delay_ms(300);
        usbDeviceConnect();

        setLoopRoutine(loop);
        radioSetRxHandler(rxHandler);
    }
    else
    {
        // This is a sender node; monitor commands from UART
        uartEnable(true, true);
        uartWatchInputBytes(1, uartRx);
    }
}
